{-# LANGUAGE ForeignFunctionInterface #-}
{-# LANGUAGE TupleSections #-}
{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE TemplateHaskell #-}
module FrontEnd.TestSuite
( withTestSuite
, testSuiteExperiment
, TestSuite
, CriticalExpression(..)
, TestSuiteType(..)
) where

import Data.Binary
import TestSuite.TestCaseBinary
import Control.Monad
import Control.Monad.Trans
import Data.Functor
import Data.IORef
import Data.Map (Map)
import Data.Sequence (Seq, (<|))
import Engine
import Exception
import Plan
import TestSuite.TestCase hiding (TestResult(..))
import TestSuite.TestCase (TestResult)
import qualified Data.Foldable as F
import qualified Data.Map as Map
import qualified Data.Sequence as Seq
import Trace.Hpc.Tix
import Util.FCLabels
import Util.HpcReflect
import qualified Data.ByteString.Lazy as B
import Codec.Compression.GZip

foreign import ccall "revertCAFs" rts_revertCAFs  :: IO ()  

data TestSuite = TS !(IORef TestSuiteData)
               | NoTS
               | RecordedTS !(IORef (Map TestExperiment TestResult))

data TestSuiteData 
  = TestSuiteData
  { criticalExpressions :: !(Seq CriticalExpression)
  , unionedTix          :: !Tix
  }

data CriticalExpression = CriticalExpression { ceExperiment :: !Experiment 
                                             , ceResult     :: !ExperimentResult
                                             , ceTix        :: !Tix
                                             }

mkLabels [''TestSuiteData]

data TestSuiteType = NoTestSuite
                   | HpcTestSuite
                   | RecordTS FilePath
  deriving (Eq, Show)

withTestSuite :: (MonadIO m, ExceptionMonad m) => TestSuiteType -> (TestSuite -> m a) -> m (a, Seq CriticalExpression)
withTestSuite NoTestSuite f = (,Seq.empty) `liftM` f NoTS
withTestSuite HpcTestSuite f = do
  ts@(TS ref) <- setup 
  res <- f ts 
  tsd <- liftIO $ readIORef ref
  liftIO $ updateTix (unionedTix tsd)
  return (res, criticalExpressions tsd)
  where
    setup = liftIO $ do
      clearTix
      initTix <- examineTix
      ref <- newIORef (emptyTestSuiteData initTix)
      clearTix
      return $! TS ref
withTestSuite (RecordTS path) f = do
  ref <- liftIO $ newIORef $! Map.empty
  res <- f (RecordedTS ref)
  endMap <- liftIO $ readIORef ref
  liftIO $ B.writeFile path . compress . encode $ endMap
  return (res, Seq.empty)

testSuiteExperiment :: TestSuite -> Experiment -> RanExperiment -> IO ()
testSuiteExperiment NoTS _ _ = return ()
testSuiteExperiment (TS ref) !exp !ranExp 
  | isResult (getExperimentResult ranExp) 
  = do
    currentTix <- examineTix
    clearTix
    rts_revertCAFs
    !otsd <- readIORef ref
    let ne = addNewExpression False
                CriticalExpression { ceExperiment = exp
                                   , ceResult = getExperimentResult ranExp
                                   , ceTix = currentTix
                                   } otsd
    writeIORef ref $! ne
  | otherwise = return ()
  where
    isResult Ok = True
    isResult (Show s) = True
    isResult (Errors _) = True
    isResult (Property _) = True
    isResult _ = False

testSuiteExperiment (RecordedTS ref) !exp !ranExp = do
  maybe (return ()) (\tc -> do
      mp <- readIORef ref
      writeIORef ref $! Map.insert (testExperiment tc) (testResult tc) mp
    ) (mkTestCase exp (getExperimentResult ranExp))

emptyTestSuiteData :: Tix -> TestSuiteData
emptyTestSuiteData initTix = TestSuiteData { criticalExpressions = Seq.empty
                                           , unionedTix = initTix
                                           }

addNewExpression :: Bool -> CriticalExpression -> TestSuiteData -> TestSuiteData
addNewExpression alwaysAdd inCe@CriticalExpression { ceExperiment = inCeExperiment
                                                   , ceResult = inCeResult
                                                   , ceTix = inCeTix 
                                                   } tsd 
  = TestSuiteData { criticalExpressions = newCE
                  , unionedTix          = newUT
                  }
  where
    isUsefulNew = inCeTix `coversNewAreaFrom` (unionedTix tsd)
    isUsefulSuperset = Seq.length (criticalExpressions tsd) /= Seq.length supersetPruned
    supersetPruned = Seq.filter ((`coversNewAreaFrom` inCeTix) . ceTix) (criticalExpressions tsd)

    newCE
      | alwaysAdd                             = inCe <| criticalExpressions tsd
      | isUsefulNew {- || isUsefulSuperset -} = inCe <| supersetPruned
      | otherwise                             = criticalExpressions tsd
    
    newUT
      | isUsefulNew || alwaysAdd = unionedTix tsd `unionTix` inCeTix
      | otherwise   = unionedTix tsd


coversBetterThan :: Integer -> Integer -> Bool
coversBetterThan x 0 = x /= 0
coversBetterThan _ _ = False

coversSameOrBetterThan :: Integer -> Integer -> Bool
coversSameOrBetterThan _ 0 = True
coversSameOrBetterThan x _ = x > 0

isSupersetOrEqualOf :: Tix -> Tix -> Bool
isSupersetOrEqualOf (Tix xs) (Tix ys) = and ssOrEq
  where
    ssOrEq = zipWith isSupersetOrEqualOfModule xs ys

    isSupersetOrEqualOfModule :: TixModule -> TixModule -> Bool
    isSupersetOrEqualOfModule (TixModule s1 h1 i1 xs) (TixModule s2 h2 i2 ys)
      | s1 == s2 && h1 == h2 && i1 == i2 = and (zipWith coversSameOrBetterThan xs ys)

isSupersetOf :: Tix -> Tix -> Bool
isSupersetOf x@(Tix xs) y@(Tix ys) = ssOrEq && or usSS
  where
    ssOrEq = isSupersetOrEqualOf x y
    usSS   = zipWith isUnsafeSupersetOfModule xs ys

    isUnsafeSupersetOfModule :: TixModule -> TixModule -> Bool
    isUnsafeSupersetOfModule (TixModule s1 h1 i1 xs) (TixModule s2 h2 i2 ys)
      | s1 == s2 && h1 == h2 && i1 == i2 = or (zipWith coversBetterThan xs ys)

coversNewAreaFrom :: Tix -> Tix -> Bool
coversNewAreaFrom (Tix xs) (Tix ys) = or cnaf
  where
    cnaf = zipWith coversNewAreaFromModule xs ys

    coversNewAreaFromModule (TixModule s1 h1 i1 xs) (TixModule s2 h2 i2 ys)
      | s1 == s2 && h1 == h2 && i1 == i2 = or (zipWith coversBetterThan xs ys)

unionTix :: Tix -> Tix -> Tix
unionTix (Tix xs) (Tix ys) = Tix zs
  where
    zs = zipWith unionTixModule xs ys

    unionTixModule :: TixModule -> TixModule -> TixModule
    unionTixModule (TixModule s1 h1 i1 xs) (TixModule s2 h2 i2 ys)
      | s1 == s2 && h1 == h2 && i1 == i2 = TixModule s1 h1 i1 (zipWith max xs ys)
