{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE NamedFieldPuns #-}
module FrontEnd.TestSuiteAnalyser
(tsaAction
) where

import Bottom
import Codec.Compression.GZip
import Control.Applicative
import Control.Monad
import Data.Binary
import Data.Map (Map)
import Data.Sequence (Seq)
import Data.Set (Set)
import Format
import FrontEnd.Options
import qualified Data.ByteString.Lazy as B
import qualified Data.Foldable as F
import qualified Data.Map as Map
import qualified Data.Sequence as Seq
import qualified Data.Set as Set
import qualified Data.Text as T
import TestSuite.TestCase


tsaAction :: () -> [String] -> IrulanOptions -> IO ()
tsaAction () [tsp1] options = do
  ts1 <- loadTestSuite tsp1

  putStrLn $ "A: " ++ tsp1

  forM_ (Map.assocs ts1) $ \(te,tr) -> do
    formatLine A te tr UNKNOWN


tsaAction () [tsp1,tsp2] options = do
  ts1 <- loadTestSuite tsp1
  ts2 <- loadTestSuite tsp2
  
  let comparedResults = buildComparison ts1 ts2

  putStrLn $ "A: " ++ tsp1
  putStrLn $ "B: " ++ tsp2
  putStrLn ""

  forM_ (Map.assocs comparedResults) (uncurry printResults)

tsaAction () _ options = do
  putStrLn "I need two test-suites to compare!"

loadTestSuite :: FilePath -> IO (Map TestExperiment TestResult)
loadTestSuite f = return . decode . decompress =<< B.readFile f 

printResults :: TestExperiment -> TestComparison -> IO ()
printResults exp (TC { moreGeneral, general, specific }) = do
  let allPass = F.all passPredicate specific
  let onePass = F.any passPredicate specific

  let gLine = formatLine moreGeneral exp general (if allPass && onePass then OK else if allPass then UNKNOWN else ERROR)

  let sLines = F.forM_ specific (\tc@(TestCase { testExperiment , testResult }) -> do
                formatLine (otherName moreGeneral) testExperiment testResult (if passPredicate tc then OK else ERROR)
                )

  case moreGeneral of
    A -> gLine >> sLines
    B -> sLines >> gLine

  putStrLn ""
  where
    passPredicate (TestCase { testExperiment, testResult }) 
      =  testResult     == general

data Modifier
  = OK
  | ERROR
  | UNKNOWN

formatLine :: TestSuiteName -> TestExperiment -> TestResult -> Modifier -> IO ()
formatLine tsName exp res mod = do
  let (eStr, eCol)
        | OK <- mod = (": ", greenWrap)
        | ERROR <- mod  = ("~ ", redWrap)
        | UNKNOWN <- mod = ("# ", yellowWrap)
      
  let line = unwords [ show tsName ++ eStr
                     , simpleOutput exp
                     , "==>" 
                     , format res
                     ]
  putStrLn $ eCol line

redWrap x = "\ESC[31m" ++ x ++ "\ESC[0m"
greenWrap x = "\ESC[32m" ++ x ++ "\ESC[0m"
yellowWrap x = "\ESC[33m" ++ x ++ "\ESC[0m"

instance Format TestResult where
  format (Errors t)   = "! " ++ T.unpack t
  format (Property p) = show p
  format (Show t)     = T.unpack t

instance Format TestExperimentConstant where
  format (TECInt     x) = show x
  format (TECChar    x) = show x
  format (TECFloat   x) = show x
  format (TECDouble  x) = show x
  format (TECInteger x) = show x


simpleOutput :: TestExperiment -> String
simpleOutput exp 
  = unwords $ [simpleOutputRoot (not . Seq.null $ teArgs exp) (teRoot exp)] ++ (F.toList $ simpleOutputArg <$> teArgs exp) 
  where
    simpleOutputArg (TEArgExperiment e)
      | TEDestructRoot {} <- teRoot e  = "(" ++ simpleOutput e ++ ")"
      | Seq.null (teArgs e)            = simpleOutput e
      | otherwise                      = "(" ++ simpleOutput e ++ ")"
    simpleOutputArg (TEArgBottom b)    = simpleBottom b
    simpleOutputArg TEArgUnconstrained = "*"

    simpleOutputRoot _ (TEConstantRoot x) = format x
    simpleOutputRoot _ (TEIdRoot x)       = T.unpack x
    simpleOutputRoot b (TEDestructRoot { teDestructCaseOf   
                                       , teDestructConstructor
                                       , teDestructVarIndex 
                                       , teDestructConstructorArity
                                       , teDestructConstructorIsInfix
                                       }) 
      | b = "(" ++ mainBit ++ ")"
      | otherwise = mainBit
      where
        mainBit = "case " ++ (simpleOutput teDestructCaseOf) ++ " of " ++ dcBit ++ " -> x"
        dcBit
          | teDestructConstructorIsInfix = head args ++ " " ++ dcString ++ " " ++ unwords (tail args)
          | otherwise = dcString ++ " " ++ unwords args
          where
            dcString = T.unpack teDestructConstructor
            args = [ if y == teDestructVarIndex then "x" else "_" | y <- [0 ..teDestructConstructorArity - 1]]

    simpleBottom b
      | b == initialBottomId         = "?"
      | otherwise                    = "?" ++ format b
