{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# OPTIONS_GHC -fno-warn-overlapping-patterns #-}
module Main where

{- With a lot of inspiration taken from cabal-install 
 - 2005 Lemmih <lemmih@gmail.com> 
 - 2006 Paolo Martini <paolo@nemail.it> 
 - 2007 Bjorn Bringert <bjorn@bringert.net> 
 - 2007 Isaac Potoczny-Jones <ijones@syntaxpolice.org> 
 - 2008-2009 Duncan Coutts <duncan@haskell.org>
 -}

import Data.Monoid
import Distribution.Simple.Command hiding (boolOpt)
import Distribution.Simple.Setup
import Engine (Strategy, fifoStrategy, unlimitedStrategy)
import Format
import FrontEnd.Options
import FrontEnd.Source
import FrontEnd.TestSuiteAnalyser
import FrontEnd.TestSuite (TestSuiteType(..))
import FrontEnd.Vis
import IdDB
import System.Environment
import System.Exit              (exitFailure)
import Util.FCLabels

main :: IO ()
main = getArgs >>= mainWorker

mainWorker :: [String] -> IO ()
mainWorker args 
  = case commandsRun irulanCommand commands args of
      CommandHelp help    -> printHelp help
      CommandList opts    -> printOptionsList opts
      CommandErrors errs  -> printErrors errs
      CommandReadyToGo (options, commandParse) -> case commandParse of
        CommandHelp help         -> printHelp help
        CommandList opts         -> printOptionsList opts
        CommandErrors errs       -> printErrors errs
        CommandReadyToGo action  -> action options
      
  where
    commands = [ sourceCommand `commandAddAction` sourceAction
               , visCommand `commandAddAction` visAction
               , tsaCommand `commandAddAction` tsaAction
               ]

    printHelp help = do
      pname <- getProgName
      putStr (help pname)

    printOptionsList = putStr . unlines

    printErrors errs = do
      putStrLn "Print errors!"
      putStr . unlines $ errs
      exitFailure

irulanCommand :: CommandUI IrulanOptions
irulanCommand 
  = CommandUI { commandName         = ""
              , commandSynopsis     = ""
              , commandUsage        = const . unlines $ 
                  [ "This is irulan by Tristan Allwood."
                  , "http://www.doc.ic.ac.uk/~tora/irulan/"
                  ]
              , commandDescription  = Nothing 
              , commandDefaultFlags = defaultOptions
              , commandOptions      = const
                  [ optionL [] ["search-strategy"]
                      "Choose the type of search strategy"
                      lRuntimePlanType
                      (choiceOpt [ ( toFlag DFS
                                   , (['d'],["depth-bounded"])
                                   , "A complete, depth-bounded strategy. Configure the bound size with --depth"
                                   )
                                 , ( toFlag DFSEvenTime
                                   , (['e'], ["dfs-even-time"])
                                   , "The depth bounded strategy, but the maximum runtime is distributed evenly amonst splits in the plan as they are encountered"
                                   )
                                 , ( toFlag Random
                                   , (['r'], ["random"])
                                   , "A random strategy.  Configure the maximum runtime per function-to-test with --functionRuntime"
                                   )
                                 , ( toFlag RestartingRandom
                                   , (['s'], ["restarting-random"])
                                   , "A random strategy, should have better heap usage at the cost of repeating run experiments"
                                   )
                                 , ( toFlag IterativeDeepening
                                   , (['a'], ["iterative-deepening"])
                                   , "A depth first, iterative deepening strategy"
                                   )
                                 ]
                      )
                  , optionL [] ["depth"]
                      "The maximum depth of the depth-bounded search strategy (if chosen)"
                      lMaximumDepth
                      (reqArg' "INT" (toFlag . fromInteger . read) 
                                     (fmap format . flagToList)
                      )
                  , optionL [] ["maximumExpressionSize"]
                      "The maximum size of expressions to evaluate for the size-bounded strategy (if chosen)"
                      lMaximumExpressionSize
                      (reqArg' "INT" (Last . Just . Just . fromInteger . read) 
                                     (const ["<maxsize>"])
                      )
                  , optionL [] ["no-maximumExpressionSize"]
                      "Disable pruning based on expression size"
                      lMaximumExpressionSize
                      (noArg (Last (Just Nothing)))
                  , optionL [] ["maximumExpressionDepth"]
                      "The maximum nesting depth of an expression to evaluate for the size-bounded strategy (if chosen)"
                      lMaximumExpressionDepth
                      (reqArg' "INT" (Last . Just . Just . fromInteger . read) 
                                     (const ["<maxdepth>"])
                      )
                  , optionL [] ["no-maximumExpressionDepth"]
                      "Disable pruning based on maximum depth of nesting in an expression"
                      lMaximumExpressionDepth
                      (noArg (Last (Just Nothing)))
                  , optionL [] ["maximumRuntime"]
                      "The maximum time for the random and time based strategies"
                      lMaximumRuntime
                      (reqArg' "SECONDS" (toFlag . read)
                                         (fmap show . flagToList)
                      )
                  , optionL [] ["expressionTime"]
                      "The maximum time before a test expression times out"
                      lMaxExpressionTime
                      (reqArg' "MICROSECONDS" (toFlag . read)
                                              (fmap show . flagToList)
                      )
                  , optionL [] ["expressionAllocs"]
                      "The maximum number of allocations a test expression can perform before it is killed"
                      lMaxExpressionAllocs
                      (reqArg' "BYTES" (toFlag . read)
                                       (fmap show . flagToList)
                      )
                  , optionL ['t'] ["trace"]
                      "Enable tracing of test expressions as they are executed"
                      lOutputExperimentTrace
                      trueArg
                  , optionL ['v'] ["debug"]
                      "Enable debugging information"
                      lOutputDebugInformation
                      trueArg
                  , optionL [] ["stats-output"]
                      "Enable output statistics to a file"
                      lOutputStatistics
                      (reqArg' "FILENAME" toFlag flagToList)
                  , optionL [] ["full-stats-output"]
                      "output of full statistics of code coverage after every step"
                      lOutputFullStatistics
                      (boolOpt [] [])
                  , optionL [] ["vco"]
                      "use of the visible constructor optimisation"
                      lUseVCO
                      (boolOpt [] [])
                  , optionL [] ["dump-ss-size"]
                      "stopping after building the support set, dumping its size"
                      lOnlyDumpSupportSetSize
                      (boolOpt [] [])
                  , optionL [] ["hpc-testsuite"]
                      "Enable output of a hpc-minimized testsuite"
                       lGenerateTestSuite
                       (noArg (Flag HpcTestSuite))
                  , optionL [] ["full-testsuite"]
                      "Enable output of a full testsuite to a file"
                       lGenerateTestSuite
                       (reqArg' "FILENAME" (toFlag . RecordTS)  
                                           (const ["<filepath>"])
                       )
                  , optionL [] ["no-testsuite"]
                       "Disable generation of a testsuite"
                       lGenerateTestSuite
                       (noArg (Flag NoTestSuite))
                  , optionL [] ["show-results"]
                       "using Show instances (if available) to visualise results"
                       lUseShowOnResult
                       (boolOpt [] [])
                  , optionL [] ["simple-output"]
                      "simple output of expressions"
                      lUseSimpleOutput
                      (boolOpt [] [])
                  , optionL [] ["profiling"]
                      "profiling of irulan usage"
                      lDoProfiling
                      (boolOpt [] [])
                  , optionL [] ["case-statements"]
                      "case statements in the plan"
                      lUseCaseStatement
                      (boolOpt [] [])
                  , optionL [] ["case-arguments"]
                      "case arguments in the plan"
                      lUseCaseArgument
                      (boolOpt [] [])
                  , optionL [] ["automatic-identifier-source"]
                      "Choose the source of automatically discovered identifiers used to build test expressions"
                      lSourceIdsFrom
                      (choiceOpt [ ( toFlag None
                                   , ([], ["no-ai"])
                                   , "Don't automatically include identifiers to use for testing."
                                   )
                                 ,
                                   ( toFlag TestModule
                                   , ([], ["ai-test"])
                                   , "When an argument of type T is needed, automatically include any identifier from the module being tested."
                                   )
                                 , ( toFlag SourceModule
                                   , ([], ["ai-source"])
                                   , "When an argument of type T is needed, lookup the declaring module for T and use identifiers from that module to make T."
                                   )
                                 ]
                      )
                  , optionL ['i'] ["include"]
                      "Choose the source modules used to provide identifiers used for testing"
                      lExplicitIdSource
                      (reqArg' "MODULENAME" (:[]) id)
                  , optionL [] ["no-cache"]
                      "disable runtime experiment value cache"
                      lCache
                      (noArg (Last (Just Nothing)))
                  , optionL [] ["fifo-cache"]
                      "enable runtime cache with a fifo-queue"
                      lCache
                      (reqArg' "MAXSIZE" (Last . Just . Just . fifoStrategy . read)
                                         (const ["<maxsize>"])
                      )
                  , optionL [] ["unlimited-cache"]
                      "enable unlimited size runtime cache"
                      lCache
                      (noArg (Last (Just (Just unlimitedStrategy))))
                  , optionL [] ["quiet"]
                      "be quiet when running experiments - don't output a progress indicator"
                      lQuiet
                      (boolOpt [] [])
                  , optionL ['p'] ["properties"]
                      "Only test property functions (called prop or prop_*)"
                      lOnlyTestProperties
                      trueArg
                  , optionL [] ["ints"]
                      "List of Int constants to use e.g [1,2,3]"
                      lIntConstants
                      (reqArg' "INT_LIST" read (return . show))
                  , optionL [] ["int-range"]
                      "Range of Int constants to use (e.g. (1,10))"
                      lIntConstants
                      (reqArg' "INT_LIST" (uncurry enumFromTo . read) (return . show))
                  , optionL [] ["chars"]
                      "List of Char constants to use e.g ['a','b','c']"
                      lCharConstants
                      (reqArg' "CHAR_LIST" read (return . show))
                  , optionL [] ["floats"]
                      "List of Float constants to use e.g [1.0,2.0,3.0]"
                      lFloatConstants
                      (reqArg' "FLOAT_LIST" read (return . show))
                  , optionL [] ["doubles"]
                      "List of Double constants to use e.g [1.0,2.0,3.0]"
                      lDoubleConstants
                      (reqArg' "DOUBLE_LIST" read (return . show))
                  , optionL [] ["integers"]
                      "List of Integer constants to use e.g [1,2,3]"
                      lIntegerConstants
                      (reqArg' "INTEGER_LIST" read (return . show))
                  , optionL [] ["dump-plan"]
                      "dumping of the plan as a .dot file"
                      lDumpPlan
                      (boolOpt [] [])
                  ]
              }

sourceCommand :: CommandUI [String]
sourceCommand 
  = CommandUI { commandName         = "source"
              , commandSynopsis     = "Compile (if necessary) and then test source files"
              , commandUsage        = const "Test the identifiers in the provided source files, compiling them if necessary"
              , commandDescription  = Nothing
              , commandDefaultFlags = []
              , commandOptions      = const 
                  [ option [] ["ghc-options"]
                      "Additional ghc options"
                      id (++)
                      (reqArg' "OPTIONS" (:[]) id)
                  ]
              } 

visCommand :: CommandUI [String]
visCommand
  = CommandUI { commandName = "vis"
              , commandSynopsis = "Interactive exploration of the Plan space"
              , commandUsage = const "Provide interactive testing of the identifiers in the provided source files, compiling them if necessary"
              , commandDescription = Nothing
              , commandDefaultFlags = []
              , commandOptions = const
                  [ option [] ["ghc-options"]
                      "Additonal ghc ptions"
                      id (++)
                      (reqArg' "OPTIONS" (:[]) id)
                  ]
              }

tsaCommand :: CommandUI ()
tsaCommand
  = CommandUI { commandName = "tsa"
              , commandSynopsis = "Compare two testsuites from different implementations (a smart diff)"
              , commandUsage = const "Show the differences between running two testsuites"
              , commandDescription = Nothing
              , commandDefaultFlags = ()
              , commandOptions = const []
              }

optionL :: SFlags -> LFlags -> Description -> (a :-> b) -> MkOptDescr (a -> b) (b -> a -> a) a1 -> OptionField a1
optionL a b d l o = option a b d (get l) (set l) o

{- urk nothing is ever simple is it? -}
instance Eq Strategy where
  _ == _ = False
