{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
module Plan.PlanM
( PlanM
, runPlanM
, PlanEnv(..)
, withNewBottomId
, askIdsForType
, askDestructPlan
, askConstantsForType
, askConstructorsForTyCon
, withExperimentResult
, lookupExperimentResult
, iExperimentToIExperimentArgument
, guardBuildCaseArguments
) where

import BasicTypes
import Bottom
import Control.Applicative
import Outputable
import Control.Monad.Reader
import Control.Monad.State.Strict
import Control.Monad.Writer
import DataCon
import Data.Map (Map)
import Data.Maybe
import IdDB
import IdDB.ConstructorGraph (DestructPlan)
import Plan.Experiment
import Plan.IExperiment
import qualified Data.Map as Map
import qualified Data.Traversable as T
import qualified Util.BiMap as BiMap
import qualified Util.UniqSet as UniqSet
import TyCon
import Type
import TysWiredIn
import Util.BiMap (BiMap)
import Util.UniqSet ( UniqSet )
import Var
import VarEnv
import Debug.Trace

newtype PlanM a = PlanM { _runPlanM :: Reader PlanEnv a }
  deriving (Functor, Monad, MonadReader PlanEnv)

runPlanM :: IdDB -> Bool -> PlanM a -> a
runPlanM iddb buildCaseArguments 
  = flip runReader newEnv . _runPlanM
  where
    newEnv = PlanEnv { iddb           = iddb
                     , knownResults   = Map.empty
                     , nextBottomId   = initialBottomId
                     , buildCaseArguments = buildCaseArguments
                     }

data PlanEnv
  = PlanEnv
  { iddb               :: !IdDB
  , knownResults       :: !(Map WeakIExperiment ExperimentResult)
  , nextBottomId       :: !BottomId
  , buildCaseArguments :: !Bool
  } 

withNewBottomId :: (BottomId -> PlanM a) -> PlanM a
withNewBottomId nxt = do
  bid <- asks nextBottomId
  local (\env -> env { nextBottomId = succ bid }) (nxt bid)

guardBuildCaseArguments :: a -> PlanM a -> PlanM a
guardBuildCaseArguments def cond = do
  bcas <- asks buildCaseArguments
  if bcas then cond else return def

askIdsForType :: InScopeSet -> Type -> PlanM [(Id, TvSubst, InScopeSet)]
askIdsForType inScopeTyVars ty 
  = asks $ idsForType inScopeTyVars ty . iddb

askConstantsForType :: Type -> PlanM [Constant]
askConstantsForType ty
  = asks $ constantsForType ty . iddb

askDestructPlan :: InScopeSet -> Type -> PlanM [((Id, TvSubst), DestructPlan, InScopeSet)]
askDestructPlan inScopeTyVars ty
  = asks $ destructPlan inScopeTyVars ty . iddb

askConstructorsForTyCon :: TyCon -> PlanM (UniqSet DataCon)
askConstructorsForTyCon tycon
  = asks $ constructorsForTyCon tycon . iddb
  {- | tycon == listTyCon = return $ UniqSet.singleton consDataCon
  | tycon == pairTyCon = return $ UniqSet.singleton $ tupleCon Boxed 2
  | otherwise = return mempty -}

lookupExperimentResult :: IExperiment -> PlanM (Maybe ExperimentResult)
lookupExperimentResult iexp = asks (lookupExperimentResult' iexp)

lookupExperimentResult' :: IExperiment -> PlanEnv -> Maybe ExperimentResult
lookupExperimentResult' oexp@(canonicaliseIExperiment -> (exp, bimap)) (knownResults -> knownResults) 
  | isSimpleCtrExperiment oexp 
      = Just Ok
  | otherwise 
      = remapExperimentResult (BiMap.inverseMap bimap) <$> Map.lookup (WeakIExperiment exp) knownResults



withExperimentResult :: IExperiment -> ExperimentResult -> PlanM a -> PlanM a
withExperimentResult exp res = local (addExperimentResult exp res)

addExperimentResult :: IExperiment -> ExperimentResult -> PlanEnv -> PlanEnv
addExperimentResult (canonicaliseIExperiment -> (exp, bimap)) er planEnv@(knownResults -> knownResults) 
  = let insExp = (remapExperimentResult (BiMap.forwardMap bimap) er)
    in insExp `seq` planEnv { knownResults = Map.insert (WeakIExperiment exp) insExp knownResults }

canonicaliseIExperiment :: IExperiment -> (IExperiment, BiMap BottomId BottomId)
canonicaliseIExperiment exp = flip evalState initialBottomId . runWriterT $ canonicaliseExperiment' exp
  where
    canonicaliseExperiment' exp = do
      args <- T.mapM canonicaliseArg (iExperimentArgs exp)
      root <- canonicaliseRoot (iExperimentRoot exp)
      return $ exp { iExperimentRoot = root, iExperimentArgs = args }

    canonicaliseArg (IArgExperiment exp) = IArgExperiment <$> canonicaliseExperiment' exp
    canonicaliseArg (IArgBottom ty obid) = do
      nbid <- lift nextBottomId 
      tell (BiMap.singleton obid nbid) 
      return $ IArgBottom ty nbid

    canonicaliseArg (IArgUnconstrained ty obid) = do
      nbid <- lift nextBottomId
      tell (BiMap.singleton obid nbid)
      return $ IArgUnconstrained ty nbid
    
    canonicaliseRoot c@(IConstantRoot _) = return c
    canonicaliseRoot r@(IIdRoot _)       = return r
    canonicaliseRoot d@(IDestructRoot { iDestructCaseOf }) = do 
      caseOf' <- canonicaliseExperiment' iDestructCaseOf
      return $ d { iDestructCaseOf = caseOf' }

    nextBottomId = State (\id -> (id, succ id))

remapExperimentResult :: Map BottomId BottomId -> ExperimentResult ->  ExperimentResult
remapExperimentResult mp (Bottoms bid) = Bottoms . fromMaybe (error "remapExperimentResult - P") . Map.lookup bid $ mp
remapExperimentResult _ er = er

iExperimentToIExperimentArgument :: IExperiment -> (IExperimentArg -> PlanM a) -> PlanM a
iExperimentToIExperimentArgument exp f = do
  let expr = IArgExperiment exp 
  -- withStatefulBottoms (relabelBottomIds expr) f
  f expr
  {-
  where
    withStatefulBottoms :: (State BottomId a) -> (a -> PlanM b) -> PlanM b
    withStatefulBottoms st f = do
       nbid <- asks nextBottomId
       let (res, nbid') = runState st nbid
       local (\env -> env { nextBottomId = nbid' }) (f res)

    relabelBottomIds :: IExperimentArg -> State BottomId IExperimentArg
    relabelBottomIds (IArgExperiment exp) = do
      args' <- T.mapM relabelBottomIds (iExperimentArgs exp)
      return $ IArgExperiment exp { iExperimentArgs = args' }
    relabelBottomIds (IArgBottom t _)     = (IArgBottom t) <$> nxtBottomId
    relabelBottomIds (IArgUnconstrained t _) = (IArgUnconstrained t) <$> nxtBottomId

    nxtBottomId = State (\id -> (id, succ id))
-}
