{-# LANGUAGE TupleSections #-}
{-# LANGUAGE ViewPatterns #-}
{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE BangPatterns #-}
module Plan.TreePlan
( TreePlan(..)
, generateTreePlan
) where

import Bottom
import Control.Applicative
import Control.Monad
import DataCon
import Data.IntMap (IntMap)
import Data.List
import Data.Map (Map)
import Data.Maybe
import Data.Monoid
import Data.Ord
import Data.Sequence (Seq, (<|), (|>), (><))
import Data.Set (Set)
import Format
import Id
import IdDB
import Outputable hiding (trace)
import Plan.Experiment
import Plan.IExperiment
import Plan.PlanM
import qualified Data.Foldable as F
import qualified Data.IntMap as IntMap
import qualified Data.Map as Map
import qualified Data.Sequence as Seq
import qualified Data.Set as Set
import qualified Data.Traversable as T
import qualified Util.UniqMap as UniqMap
import qualified Util.UniqSet as UniqSet
import Type
import TysWiredIn (unitTy)
import Util
import Util.GHC
import Util.UniqMap (UniqMap)
import Util.UniqSet (UniqSet)
import Var
import VarEnv

data TreePlan
  = TPStep
  { tpExperiment  :: !Experiment
  , tpOnOk        :: !TreePlan
  , tpOnDataCon   :: !(UniqMap DataCon TreePlan)
  , tpOnBottom    :: !(Map BottomId TreePlan)
  }
  | TPSplit 
  { tpChildren    :: !(Seq TreePlan)
  }

stop :: TreePlan
stop = TPSplit { tpChildren = Seq.empty }

smartSplit :: Seq TreePlan -> TreePlan
smartSplit (Seq.filter (not . isStopPlan) -> plan) 
  = case Seq.length plan of
      0 -> stop
      1 -> plan `Seq.index` 0
      _ -> TPSplit { tpChildren = plan }

isStopPlan :: TreePlan -> Bool
isStopPlan TPSplit { tpChildren = (Seq.null -> isStop) } = isStop
isStopPlan _ = False

type Instantiate = IExperiment -> TvSubst -> InScopeSet -> PlanM (TreePlan, UniqMap DataCon TreePlan)

data PlanContext
  = PlanContext
  { instantiate   :: !Instantiate
  , targetType    :: !Type
  , inScopeTyVars :: !InScopeSet
  , usedIds       :: !(Set Id)
  }

generateTreePlan :: IdDB -> Bool -> Bool -> TreePlan
generateTreePlan !iddb !addCaseArguments !destructResults = smartSplit . Seq.fromList . generateTreePlan' $ ids
  where
    generateTreePlan' = runPlanM iddb addCaseArguments . mapM buildAPlan
    ids = functionsToTest iddb
    
    buildAPlan id = buildTreePlan (makeInitialExperiment id emptyTvSubst) newContext
      where
        newContext = PlanContext { instantiate   = if destructResults 
                                                    then tryToDestructTreePlan 
                                                    else \_ _ _ -> return (mempty, mempty)
                                 , targetType    = snd . splitFunTys $ targetRawType
                                 , inScopeTyVars = extendInScopeSetList emptyInScopeSet targetTyVars
                                 , usedIds       = Set.singleton id
                                 }

        (targetTyVars, targetRawType) = splitForAllTys . varType $ id
    

buildTreePlan :: IExperiment -> PlanContext -> PlanM TreePlan
buildTreePlan !exp (dat@PlanContext { instantiate   = instantiate
                                    , targetType    = targetType
                                    , inScopeTyVars = inScopeTyVars
                                    , usedIds       = usedIds
                                    })
  = do
    mExpRes <- lookupExperimentResult exp

    let uncachedVersion 
          = do
              (okPlans, dataConPlans) <- withExperimentResult exp Ok mkOkPlan
              bottomPlans <- T.sequence $ Map.mapWithKey (\bid -> withExperimentResult exp (Bottoms bid)) (bottomPlanMs dat exp)
              return $! TPStep { tpExperiment = iExperimentToExperiment exp
                               , tpOnOk       = okPlans
                               , tpOnDataCon  = dataConPlans
                               , tpOnBottom   = bottomPlans
                               }

    case mExpRes of
      Just Ok                 -> do
        (okPs, dcPs) <- mkOkPlan
        if (UniqMap.null dcPs) 
          then return okPs
          else uncachedVersion
      Just (Bottoms bid)      -> do
        fromMaybe (return stop) $ Map.lookup bid (bottomPlanMs dat exp)
      Nothing -> do
        uncachedVersion

  where
    mkOkPlan :: PlanM (TreePlan, UniqMap DataCon TreePlan)
    mkOkPlan
      | Just match <- maybeMatch = instantiate exp match inScopeTyVars
      -- | trace ("No match: " ++ format exp ++ " :: " ++ show expType ++ " vs " ++ show targetType) False = undefined
      | otherwise                = withNewBottomId $ \bid -> case addBottomArg exp bid of
                                                              Nothing   -> return (mempty, mempty)
                                                              Just exp' -> (,mempty) <$> buildTreePlan exp' dat
      where
        maybeMatch = ut
        expType = iExperimentType exp
        ut = unifyTypes expType targetType
    
bottomPlanMs :: PlanContext -> IExperiment -> Map BottomId (PlanM TreePlan)
bottomPlanMs (dat@PlanContext { inScopeTyVars, usedIds, targetType }) exp = Map.mapWithKey makeBottomPlansForAType (bottomIdTypes exp)
  where
    makeBottomPlansForAType :: BottomId -> Type -> PlanM TreePlan
    makeBottomPlansForAType bid ty
      | isTyVarTy ty = buildTreePlan (substBottomId bid (IArgUnconstrained (getTyVar "Impossible!" ty) bid) exp) dat
    makeBottomPlansForAType bid ty = makePlanForAType inScopeTyVars ty newInstantiate True
      where
        newInstantiate :: Instantiate
        newInstantiate newExp newSubst newTyVars
          = iExperimentToIExperimentArgument newExp $ \ta ->
              let substExp = substExpFn newSubst
                           . substBottomId bid ta
                           $ exp
              in (,mempty) <$> buildTreePlan substExp dat { targetType    = substTy newSubst targetType
                                                          , inScopeTyVars = newTyVars
                                                          }

    makePlanForAType :: InScopeSet -> Type -> Instantiate -> Bool -> PlanM TreePlan
    makePlanForAType inScopeTyVars ty newInstantiate useConstructors = do
      normalPlans <- do
        possibles <- Seq.fromList . (if useConstructors then id else filter (not . isConLikeId . fst3)) 
                   . sortBy (comparing (idNumArguments . fst3)) <$> askIdsForType inScopeTyVars ty
        (T.mapM (processSource (F.any ((`Set.notMember` usedIds) . fst3) possibles)) possibles)

      caseArgPlans <- guardBuildCaseArguments Seq.empty $ do
                        destructPlans <- askDestructPlan inScopeTyVars ty

                        fmap Seq.fromList . forM destructPlans $ \((id, subst), destructPlan, inScopeSet) -> do
                          let initialExperiment = makeInitialExperiment id subst
                          TPSplit . Seq.singleton <$> buildTreePlan initialExperiment PlanContext { instantiate   = processDestructPlan destructPlan
                                                                                                  , targetType    = snd . splitFunTys . iExperimentType $ initialExperiment
                                                                                                  , inScopeTyVars = inScopeSet
                                                                                                  , usedIds       = Set.singleton id
                                                                                                  }
      constantPlans <- do
        possibles <- askConstantsForType ty
        fmap Seq.fromList . forM possibles $ \const -> do
          (cont, _) <- newInstantiate (makeConstantExperiment const) emptyTvSubst inScopeTyVars
          return cont
                        
      return $ smartSplit (normalPlans >< caseArgPlans >< constantPlans)

      where
        processSource :: Bool -> (Id, TvSubst, InScopeSet) -> PlanM TreePlan
        processSource hasBaseCases (source, match, iss)
          | source `Set.member` usedIds = if hasBaseCases 
                                            then TPSplit . Seq.singleton <$> nextStep 
                                            else return stop
          | otherwise                   = nextStep
          where
            nextStep = buildTreePlan (makeInitialExperiment source match) 
                                     dat { instantiate = newInstantiate
                                         , targetType = ty
                                         , inScopeTyVars = iss
                                         , usedIds = Set.insert source usedIds
                                         }

        processDestructPlan :: DestructPlan -> Instantiate
        processDestructPlan dp iexp subst iss  = do
          (acceptingHere, acceptingNext) <- acceptingBits (accepting dp)
          stepNexts                      <- stepNextBits (step dp)

          let combinedNexts = acceptingNext `mappend` stepNexts

          return (acceptingHere, combinedNexts)
          where
            acceptingBits True  = newInstantiate iexp subst iss
            acceptingBits False = return (stop, mempty)
            
            stepNextBits Nothing    = return mempty
            stepNextBits (Just mp)  = fmap mconcat . mapM (uncurry mkDestructCtrStep) . UniqMap.elems $ mp

            (tyCon, argTys) = splitTyConApp $ iExperimentType iexp

            mkDestructCtrStep :: DataCon -> IntMap DestructPlan -> PlanM (UniqMap DataCon TreePlan)
            mkDestructCtrStep dc = fmap mconcat . mapM (uncurry mkDestructStep) . IntMap.toList
              where
                dataConArgTys = dataConInstArgTys dc argTys

                mkDestructStep :: Int -> DestructPlan -> PlanM (UniqMap DataCon TreePlan)
                mkDestructStep idx nextDP = do
                  let currentVT = dataConArgTys !! idx
                  let die = destructIExperiment iexp dc idx currentVT
                  (onOk, onDataCon) <- withExperimentResult die Ok (processDestructPlan nextDP die subst iss)
                  onBottom <- T.sequence $ Map.mapWithKey (\bid -> withExperimentResult die (Bottoms bid)) 
                                                          (bottomPlanMs (dat { instantiate   = processDestructPlan nextDP
                                                          , targetType    = currentVT
                                                          , inScopeTyVars = iss
                                                          }) die)

                  return $ UniqMap.singleton dc $ TPStep { tpExperiment = iExperimentToExperiment die
                                                         , tpOnOk       = onOk
                                                         , tpOnDataCon  = onDataCon
                                                         , tpOnBottom   = onBottom
                                                         }
{-
        processDataCon :: (DataCon, Type, Int, InScopeSet) -> PlanM TreePlan
        processDataCon (dc, dataConTargetType, varIdx, iss) = do
          -- () <- trace ("processDataCon: " ++ showPpr dc ++ " - " ++ show dataConTargetType ++ " origTy: " ++ show ty) (return ())
          TPSplit <$> makePlansForAType iss dataConTargetType pdcInstantiate False
          where
            pdcInstantiate :: IExperiment -> TvSubst -> InScopeSet -> PlanM (Either TreePlan (UniqMap DataCon (Seq TreePlan)))
            pdcInstantiate newExp newSubst newTyVars
              = do
                let varTy = substTy newSubst ty
                let exp' = destructIExperiment newExp dc varIdx varTy
                -- () <- trace ("pdcInstantiate: varTy:" ++ show varTy ++ " dcTT:" ++ show dataConTargetType ++ "exp': " ++ format exp') (return ())
                p' <- newInstantiate exp' newSubst newTyVars

                return $ Right $ UniqMap.singleton dc $ Seq.singleton $ TPStep { tpExperiment = iExperimentToExperiment exp'
                                                                               , tpOnOk       = p'
                                                                               , tpOnBottom   = Map.empty
                                                                               }
-}

instL :: (IExperiment -> TvSubst -> InScopeSet -> PlanM TreePlan) 
      -> Instantiate
instL !f !a1 !a2 !a3 = (,mempty) <$> f a1 a2 a3

instR :: (IExperiment -> TvSubst -> InScopeSet -> PlanM (UniqMap DataCon TreePlan))
      -> Instantiate 
instR !f !a1 !a2 !a3 = (mempty,) <$> f a1 a2 a3

tryToDestructTreePlan :: Instantiate
tryToDestructTreePlan iexp subst iss
  | (argTys, resultTy) <- splitFunTys ty
  , not . null $ argTys                                
    = (,mempty) <$> buildTreePlan iexp (PlanContext { instantiate = tryToDestructTreePlan
                                                    , targetType = resultTy
                                                    , inScopeTyVars = iss
                                                    , usedIds = Set.empty
                                                    })
  where
    ty = iExperimentType iexp

tryToDestructTreePlan iexp subst iss
  | Just (tyCon, tyArgs)     <- splitTyConApp_maybe ty = do
    dcs <- UniqSet.toList <$> askConstructorsForTyCon tyCon
    theMap <- mconcat <$> mapM (mkDataConMap tyArgs) dcs
    return $! (mempty,) $! theMap
  where
    ty = iExperimentType iexp
    mkDataConMap :: [Type] -> DataCon -> PlanM (UniqMap DataCon TreePlan)
    mkDataConMap tyArgs dc = UniqMap.singleton dc <$> destructDataCon tyArgs dc

    destructDataCon :: [Type] -> DataCon -> PlanM TreePlan
    destructDataCon tyArgs dc = (smartSplit . Seq.fromList) <$> (sequence . catMaybes . zipWith mkOne [0..] . map (substTy subst) . dataConRepArgTys $ dc)
      where
        subst = zipOpenTvSubst (dataConUnivTyVars dc) tyArgs

        mkOne :: Int -> Type -> Maybe (PlanM TreePlan)
        mkOne i t
          | isUnLiftedType t = Nothing
          | otherwise = Just $ do
            (onOk, onDataCon) <- withExperimentResult theExperiment Ok (tryToDestructTreePlan theExperiment subst iss)
            let dat = PlanContext { instantiate = tryToDestructTreePlan
                                  , targetType = t
                                  , inScopeTyVars = iss
                                  , usedIds = Set.empty 
                                  }
            bottomPlans <- T.sequence $ Map.mapWithKey (\bid -> withExperimentResult theExperiment (Bottoms bid)) (bottomPlanMs dat theExperiment)
            return $ TPStep { tpExperiment = iExperimentToExperiment theExperiment
                            , tpOnOk       = onOk
                            , tpOnDataCon  = onDataCon
                            , tpOnBottom   = bottomPlans
                            }
            where
              theExperiment = destructIExperiment iexp dc i t

tryToDestructTreePlan _ _ _ = return $! (mempty,mempty)

instance Monoid TreePlan where
  mempty = stop
  tpL `mappend` tpR = smartSplit (Seq.fromList [tpL, tpR])
